#!/usr/bin/env python3
"""


"""
# ruff: noqa:

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
from weakref import ref
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
# from dataclasses import InitVar, dataclass, field
# from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
default_log_colours : Final[dict[int, tuple[str, str]]] = {
    logmod.DEBUG    : ("fg", "grey"),
    logmod.INFO     : ("fg", "blue"),
    logmod.WARNING  : ("fg", "yellow"),
    logmod.ERROR    : ("fg", "red"),
    logmod.CRITICAL : ("fg", "red"),
}

default_colour_mapping : Final[dict[str, tuple[str,str]]] = {
    "blue"           : ("fg", "blue"),
    "cyan"           : ("fg", "cyan"),
    "green"          : ("fg", "green"),
    "magenta"        : ("fg", "magenta"),
    "red"            : ("fg", "red"),
    "yellow"         : ("fg", "yellow"),
    "bg_blue"        : ("bg", "blue"),
    "bg_cyan"        : ("bg", "cyan"),
    "bg_green"       : ("bg", "green"),
    "bg_magenta"     : ("bg", "magenta"),
    "bg_red"         : ("bg", "red"),
    "bg_yellow"      : ("bg", "yellow"),
    "bold"           : ("ef", "bold"),
    "underline"      : ("ef", "u"),
    "italic"         : ("ef", "italic"),
    "RESET"          : ("rs", "all"),
}
# Body:
