"""
This module provides some miscellaneous mixin classes.

SubAnnotate_m and SubRegistry_m allow for creating annotated subclasses.
(eg: Strang[int])

Human_m is for rounding time and formatting large numbers easier.

PathManip_m is a simple toolbox for some path operations I keep using.

Zipper_m is for zipfile operations.

"""

